/*
 * Generic serial functions  (compatible with PalmOS>=3.3)
 *
 *  $Id: newserial.c,v 1.4 2002/04/02 19:49:18 ldrolez Exp $ 
 *
 *  Copyright (C) 2000 Ludovic Drolez
 *
 *   This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
 *
 */

#include <PalmOS.h>

#include "palmwac.h"
#include "main_rsrc.h"

#define NEWSERIAL_PORTNAME "ircm"

union portname {
	Char name[4];
	UInt32 port;
};

static UInt16  portID;
static Char *portnamestr = NEWSERIAL_PORTNAME;

Boolean CheckForNewSerialManager(void)
{
   Err err;
   UInt32 value;
   err = FtrGet(sysFileCSerialMgr, sysFtrNewSerialPresent, &value);
   return ((!err) && (value));
}

Err     SerialOpen (UInt32 baud)
{
   Err     err;
   union portname port;
   UInt16 i;
   if(!CheckForNewSerialManager()) {
      FrmAlert(NoNewSerialManager);
      return -1;
   }
   if (gPref.debug) {   
      Debug ("New serial manager: opening " NEWSERIAL_PORTNAME);
   }
   for(i=0; i<4; i++)
      port.name[i] = portnamestr[i];
   err = SrmOpen (port.port, baud, &portID);
   if (err != 0) {
      if (err == serErrAlreadyOpen) {
	 FrmAlert(SerialInUseAlert);
      } else {
	 FrmAlert(CantopenserialAlert);
	 return true;
      }
   }
   return err;
}

void    SerialClose (void)
{
   SrmClose (portID);
}

Err     SerialTx (void *buf, UInt32 count, UInt32 *nbytes)
{
   Err err;
   *nbytes = SrmSend (portID, buf, count, &err);
   return err;
}

Err     SerialRx (void *buf, UInt32 count, UInt32 timeout, UInt32 *nbytes)
{
   Err err;
   *nbytes = SrmReceive (portID, buf, count, timeout, &err);
   return err;
}

Err     SerialRxCheck (UInt32 *nbytes)
{
   return SrmReceiveCheck (portID, nbytes);
}

Err     SerialTxString(Char *s)
{
   UInt32 nbytes;
   return SerialTx(s, StrLen (s), &nbytes);
}
